DEFAULT_NSW_SEARCH_PROMPT = """
- Hãy chuẩn hóa từ {nsw} thành dạng chuẩn

- Giải nghĩa từ {nsw} với từng dạng chuẩn

- Thêm một số biến thể viết tắt cho mỗi dạng chuẩn của {nsw}  (nếu có)

- Đặt một câu ví dụ với từ không chuẩn {nsw}.

Kết quả trả về ở dạng HTML.
"""

DEFAULT_LEXNORM_PROMPT = """examples"""