from .hate_speech_detection import HateSpeechDetection
from .hate_speech_span_detection import HateSpeechSpanDetection
from .spam_review_detection import SpamReviewDetection
from .emotion_recognition import EmotionRecognition
from .aspect_sentiment_analysis import AspectSentimentAnalysis

__all__ = [
    "HateSpeechDetection",
    "HateSpeechSpanDetection",
    "SpamReviewDetection",
    "EmotionRecognition",
    "AspectSentimentAnalysis"
]

